//
//  Devices.m
//  tys-terminal-app
//
//  Created by AR'S Chino Mac 1 on 10/31/16.
//  Copyright © 2016 Taiyo Yuden. All rights reserved.
//

#import "Devices.h"
#import "CustomCells.h"
#import "Terminal.h"

@interface DeviceSelection ()
@property (strong, nonatomic) IBOutlet DeviceInfo *cbDeviceEntry;
@property (strong, nonatomic) IBOutlet UITableView *tableView;
@property (strong) NSMutableArray *deviceList;
@property (strong) NSMutableArray *rssiList;
@property (nonatomic, strong) CBCentralManager *cbManager;
@property (nonatomic, strong) CBPeripheral *connected_peripheral;
@end

@implementation DeviceSelection


NSMutableArray *deviceList, *rssiList;
CBCentralManager *cbManager;
CBPeripheral *connected_peripheral;
NSArray *services;

#pragma mark UIViewController

/**
*   UIViewController: called when this view is loaded into the application. Initialises
*   Bluetooth and member variables
*/
- (void)viewDidLoad {
    [super viewDidLoad];

    deviceList = [NSMutableArray array];
    rssiList = [NSMutableArray array];
    services = @[[CBUUID UUIDWithString:TAIYOYUDEN_SERVICE_UUID]];

    // CoreBluetooth manager initialisation
    cbManager = [[CBCentralManager alloc] initWithDelegate:self queue:nil];
    self.cbManager = cbManager;

    // Register this view as a delegate for tableview events
    _tableView.dataSource=self;
    _tableView.delegate=self;

    }

/**
*   UIViewController: called when the view is displayed on screen. Disconnects any
*   current Bluetooth peripheral connections within the application
*/
- (void) viewWillAppear:(BOOL)animated {
    NSLog(@"viewWillAppear");
    if (connected_peripheral != NULL){
        // Disconnect a peripheral if connected
        [cbManager cancelPeripheralConnection:connected_peripheral];
        connected_peripheral = NULL;
        NSLog(@"Disconnected peripheral");
    }
    [self refreshView];
}

/**
*   UIViewController: Delegate function called prior to scene change. Attaches the
*   CBPeripheral instance of the connected device to send to the terminal application.
*/
- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender {
    Terminal *destination = segue.destinationViewController;
        destination.peripheral = connected_peripheral;
}

- (void)didReceiveMemoryWarning {
    [super didReceiveMemoryWarning];
    // Dispose of any resources that can be recreated.
}

/*
*   Cleanup function. Removes this application from any notifications it has been
*   registered to observe
*/
- (void) dealloc {
    NSLog(@"dealloc");
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

#pragma mark UITableView

/**
*   UITableView: Get the number of rows in the current table section
*   @param section Section index to search
*   @return Number of rows in the specified section
*/
- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    return [deviceList count];
}

/**
*   UITableView: Returns a cell populated with information from a discovered peripheral to
*   populate the UITableView. Cell style is based on the 'DeviceInfo' custom class
*   @param indexPath Identifier for the table row
*   @return UITableViewCell object containing peripheral information
*/
-(UITableViewCell*)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    static NSString *cellID = @"cbDevice";
    DeviceInfo *cell = (DeviceInfo *)[tableView dequeueReusableCellWithIdentifier: cellID];

    // Create new cell if necessary
    if (cell == nil)
    {
        NSArray *nib = [[NSBundle mainBundle] loadNibNamed:cellID owner:self options:nil];
        cell = [nib objectAtIndex:0];
    }

    // Populate cell
    CBPeripheral *device = [deviceList objectAtIndex:indexPath.row];
    cell.name.text = [device name];
    NSString *ident = [[device identifier] UUIDString];
    cell.ident.text = [ident substringFromIndex:[ident length]-10];
    cell.rssi.text = [NSString stringWithFormat:@"%@ dBm",[rssiList objectAtIndex:indexPath.row]];

    return cell;
}

/**
*   UITableView: Delegate function called when a row is selected by the user. Begins
*   connection process to selected device
*   @param indexPath Row selected by the user
*/
- (void)tableView:(UITableView *)tableView didSelectRowAtIndexPath:(NSIndexPath *)indexPath
{
    [tableView deselectRowAtIndexPath:indexPath animated:NO];
    CBPeripheral *device = [deviceList objectAtIndex:[indexPath row]];

    [cbManager connectPeripheral:device options:nil];

}

#pragma mark - CBCentralManagerDelegate

/**
*   CBCentralManager: Delegate function called when device successfully connected. Begins
*   service discovery for this peripheral device.
*   @param peripheral The device that was connected.
*/
- (void)centralManager:(CBCentralManager *)central didConnectPeripheral:(CBPeripheral *)peripheral
{
    NSLog(@"Connected to %@",[peripheral name]);
    // Register for CBPeripheralDelegate events
    [peripheral setDelegate:self];
    // Begin service discovery
    [peripheral discoverServices:nil];
}

/**
*   CBCentralManager: Delegate function called when peripheral is disconnected. Forces
*   the Terminal application scene to return to the device selection scene.
*   @param peripheral The device that was disconnected.
*/
- (void) centralManager:(CBCentralManager *)central didDisconnectPeripheral:(CBPeripheral *)peripheral error:(NSError *)error {

    NSLog(@"Disconnected");
    // Move back to the parent View (Terminal -> Devices)
    [self.navigationController popViewControllerAnimated:YES];
}

/**
*   CBCentralManager: Delegate function called when an advertising packet is received. Adds
*   the device name and RSSI to the current list of discovered devices.
*   @param peripheral The device that sent the advertising packet.
*/
- (void)centralManager:(CBCentralManager *)central didDiscoverPeripheral:(CBPeripheral *)peripheral advertisementData:(NSDictionary *)advertisementData RSSI:(NSNumber *)RSSI
{
    // Invalid adv. packet
    if ([peripheral name].length <= 0) return;

    // Append new device to list of known devices.
    if ([deviceList indexOfObject:peripheral] == NSNotFound){
        [deviceList addObject:peripheral];
        [rssiList addObject:RSSI];
    }
    // Refresh the UITableView with the new device
    [self.tableView reloadData];

}

/**
*   CBCentralManager: Delegate function called whenever the Bluetooth manager changes state.
**/
- (void)centralManagerDidUpdateState:(CBCentralManager *)central
{
    // Determine the state of the peripheral
    if ([central state] == CBManagerStatePoweredOff) {
        NSLog(@"CoreBluetooth BLE hardware is powered off");
        [cbManager stopScan];
    }
    else if ([central state] == CBManagerStatePoweredOn) {
        NSLog(@"CoreBluetooth BLE hardware is powered on and ready");

        [self refreshView];
    }
    else if ([central state] == CBManagerStateUnauthorized) {
        NSLog(@"CoreBluetooth BLE state is unauthorized");
    }
    else if ([central state] == CBManagerStateUnknown) {
        NSLog(@"CoreBluetooth BLE state is unknown");
    }
    else if ([central state] == CBManagerStateUnsupported) {
        NSLog(@"CoreBluetooth BLE hardware is unsupported on this platform");
    }
}

#pragma mark - CBPeripheralDelegate

/**
*   CBPeripheral: Delegate function called when service discovery is finished. Checks
*   if the device contains a service UUID matching the Taiyo Yuden service UUID and triggers
*   Characteristic discovery if valid.
*/
- (void)peripheral:(CBPeripheral *)peripheral didDiscoverServices:(NSError *)error
{
    for (CBService *service in [peripheral services]) {
        NSLog(@"Discovered service: %@", [service UUID]);

        // Does it match the TYD service UUID?
        if ([[service UUID] isEqual:[CBUUID UUIDWithString:TAIYOYUDEN_SERVICE_UUID]]){
            NSLog(@"TYD device found");
            // Valid service UUID, check for TX/RX characteristics
            [peripheral discoverCharacteristics:nil forService:service];
        }
    }
}

/**
*   CBPeripheral: Delegate function called when Characteristic discovery is finished. Checks
*   if the device contains UUIDs matching the Taiyo Yuden TX/RX. If the device contains both,
*   the Terminal application scene is launched.
*   @param service Service hosting the discovered characteristics.
*/
- (void)peripheral:(CBPeripheral *)peripheral didDiscoverCharacteristicsForService:(CBService *)service error:(NSError *)error
{
    int svc_count=0;
    for (CBCharacteristic *ch in [service characteristics]){
        NSLog(@"Characteristic: %@", [ch UUID]);
        // Is this one of the UUIDs we are looking for?
        if ([[ch UUID] isEqual:[CBUUID UUIDWithString:TAIYOYUDEN_TXCHAR_UUID]] || [[ch UUID] isEqual:[CBUUID UUIDWithString:TAIYOYUDEN_RXCHAR_UUID]] ){
            // We need both before we can continue
            svc_count++;
        }
    }
    if (svc_count == 2){
        // Vallid device, launch terminal application
        connected_peripheral = peripheral;
        [self performSegueWithIdentifier:@"TerminalSegue" sender:@"Devices"];
    } else {
        // Invalid device, disconnect from this device.
        NSLog(@"Not running TYSA-B application");
        [cbManager cancelPeripheralConnection:peripheral];
    }
}

#pragma mark User Functions

/**
*   Clears all currently discovered devices and the UITableView and restarts scanning.
*/
- (void) refreshView {
    NSLog(@"refreshView");
    [deviceList removeAllObjects];
    [rssiList removeAllObjects];
    [_tableView reloadData];

    if ([cbManager state] == CBManagerStatePoweredOn){
        [cbManager scanForPeripheralsWithServices:nil options:@{CBCentralManagerScanOptionAllowDuplicatesKey:@YES}];
    }
}


@end
