//
//  Terminal.m
//  tys-terminal-app
//
//  Created by AR'S Chino Mac 1 on 11/9/16.
//  Copyright © 2016 Taiyo Yuden. All rights reserved.
//

#include "CustomCells.h"
#include "Terminal.h"
#include "Devices.h"
#import  "QuartzCore/QuartzCore.h"

@interface Terminal()

@property (strong, nonatomic) IBOutlet UIButton *Send;
@property (strong, nonatomic) IBOutlet UIBarButtonItem *OPT;
@property (strong, nonatomic) IBOutlet UIButton *EOL;
@property (strong, nonatomic) IBOutlet UIButton *Format;
@property (strong, nonatomic) IBOutlet UIButton *Clear;

@property (strong, nonatomic) IBOutlet UITextField *txt_message;
@property (strong, nonatomic) IBOutlet NSLayoutConstraint *ControlsViewBtmConstraint;

@property (strong, nonatomic) IBOutlet UITableView *messageTable;
@property (strong) NSMutableArray *messages;


@end

@implementation Terminal


NSArray *uuids;
CBPeripheral *peripheral;
CBService *service;
CBCharacteristic *TxChar, *RxChar;
NSMutableArray *messages, *direction;

eol_t eol = EOL_NONE;
format_t format = Fmt_Ascii;

#pragma mark UIViewController

/**
*   UIViewController: called when this view is loaded into the application. Initialises
*   member variables, subscribes to CBPeripheralDelegate events and notifications from
*   the TX CBCharacteristic
*/
- (void) viewDidLoad {
    [super viewDidLoad];
    
    [_EOL setTitle:[NSString stringWithFormat:@"EOL:%@",eol2str(eol)] forState:UIControlStateNormal];
    [_Format setTitle:[NSString stringWithFormat:@"%@",format2str(format)] forState:UIControlStateNormal];
    messages = [NSMutableArray array];
    direction = [NSMutableArray array];
    
    uuids=@[[CBUUID UUIDWithString:TAIYOYUDEN_SERVICE_UUID],
            [CBUUID UUIDWithString:TAIYOYUDEN_TXCHAR_UUID],
            [CBUUID UUIDWithString:TAIYOYUDEN_RXCHAR_UUID]];

    _messageTable.dataSource=self;
    _messageTable.delegate=self;
    _messageTable.estimatedRowHeight = 70;
    _messageTable.rowHeight = UITableViewAutomaticDimension;
    _txt_message.delegate = self;
    
    peripheral = _peripheral;
    [peripheral setDelegate:self];
    
    // Find the TYD service
    NSPredicate *predicate = [NSPredicate predicateWithFormat:@"UUID == %@", uuids[0]];
    service = [[[peripheral services] filteredArrayUsingPredicate:predicate] objectAtIndex:0];
    
    // Link the characteristics
    predicate = [NSPredicate predicateWithFormat:@"UUID == %@", uuids[1]];
    TxChar = [[[service characteristics] filteredArrayUsingPredicate:predicate] objectAtIndex:0];
    
    predicate = [NSPredicate predicateWithFormat:@"UUID == %@", uuids[2]];
    RxChar = [[[service characteristics] filteredArrayUsingPredicate:predicate] objectAtIndex:0];
    
    // Subscribe to notifications
    [peripheral setNotifyValue:YES forCharacteristic:RxChar];
    
}

/**
*   UIViewController: called when the view is displayed on screen. Registers for on-screen
*   keyboard events (show and hide).
*/
- (void) viewWillAppear:(BOOL)animated {

    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(keyboardWillShow:)
                                                 name:UIKeyboardWillShowNotification
                                               object:nil];
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(keyboardWillHide:)
                                                 name:UIKeyboardWillHideNotification
                                               object:nil];
}

/**
*   UIViewController: called when the view leaves the screen. Unregisters from on-screen
*   keyboard events.
*/
- (void) viewWillDisappear:(BOOL)animated {
    [[NSNotificationCenter defaultCenter] removeObserver:self];
}

#pragma mark Keyboard events

/**
*   Notification received when on-screen keyboard is shown. Slides the layout up
*   so the keyboard does not obscure the application.
*/
- (void) keyboardWillShow:(NSNotification *) notification {
    NSLog(@"Show");
    CGRect keyboardEndFrame = [notification.userInfo[UIKeyboardFrameEndUserInfoKey] CGRectValue];
    _ControlsViewBtmConstraint.constant = -1 * keyboardEndFrame.size.height + 35;
}

/**
*   Notification received when on-screen keyboard is hidden. Slides the layout
*   back to its original position
*/
- (void) keyboardWillHide:(NSNotification *) notification {
    _ControlsViewBtmConstraint.constant = 35;
}

#pragma mark - UITableView

/**
*   UITableView: Get the number of rows in the current table section
*   @param section Section index to search
*   @return Number of rows in the specified section
*/
- (NSInteger)tableView:(UITableView *)tableView numberOfRowsInSection:(NSInteger)section
{
    return [messages count];
}

/**
*   UITableView: Returns a cell populated with the associated message to
*   populate the UITableView. Cell style is based on the 'MsgEntry' custom class
*   @param indexPath Identifier for the table row
*   @return UITableViewCell object containing message
*/
-(MsgEntry*)tableView:(UITableView *)tableView cellForRowAtIndexPath:(NSIndexPath *)indexPath
{
    static NSString *cellID = @"MsgEntry";
    MsgEntry *cell = (MsgEntry *)[tableView dequeueReusableCellWithIdentifier: cellID];
    
    cell.msg.text = [messages objectAtIndex:indexPath.row];
    // Set bold font if message is TX, plain if RX
    if ([[direction objectAtIndex:indexPath.row] isEqual: TxChar]){
        cell.msg.font = [UIFont boldSystemFontOfSize:17.0];
    } else {
        cell.msg.font = [UIFont systemFontOfSize:17.0];
    }
    
    return cell;
}

#pragma mark CBPeripheral

/**
*   CBPeripheral: Delegate function called when chracteristic notifications are set.
*   @param characteristic The characteristic requested for notifications.
*   @param error Error message if notification subccription failed.
*/
- (void)peripheral:(CBPeripheral *)peripheral
didUpdateNotificationStateForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error
{
    if (error) {
        NSLog(@"Error changing notification state: %@",
              [error localizedDescription]);
    } else {
        NSLog(@"Subscribed to characteristic");
    }
}

/**
*   CBPeripheral: Delegate function called when characteristic (RX) value is changed. Signals
*   that a message was received from the peripheral device.
*   @param characteristic characteristic which received the new value.
*/
- (void)peripheral:(CBPeripheral *)peripheral
didUpdateValueForCharacteristic:(CBCharacteristic *)characteristic error:(NSError *)error{
    // Get data from characteristic
    NSData* message = [characteristic value];
    // Convert to appropriate format
    switch (format){
        case Fmt_Hex:
            message = [self hex2string:message];
            break;
        case Fmt_Ascii:
            break;
    }

    // Convert to printable string and print to screen
    NSString *string = [[NSString alloc] initWithData:message encoding:NSASCIIStringEncoding];

    [self printMessageToScreen:string characteristic:RxChar];
}

#pragma mark user functions

/**
*   Transmits a string of characters to the connected peripheral using the TX characteristic.
*   Characters are transmitted in blocks conforming to maximum BLE packet size. The string
*   will be converted if a non-Ascii format is selected.
*   @param string Message input by user
*   @return True on success, False on failure
*/
- (BOOL)writeMessageToPeripheral:(NSString *)string {

    // Convert string to transmittable byte format
    NSData *data = [string dataUsingEncoding:NSASCIIStringEncoding];

    // Convert if non-ascii format selected
    switch (format){
        case Fmt_Ascii:
            break;
        case Fmt_Hex:
            data=[self ascii2hex:data];
            // If this function returns null we cannot continue
            if (data == NULL)
                return false;
    }

    NSInteger chunk_sz = 20, sent=0, to_send=[data length];

    // Send packets until less than a full chunk remains
    while ([data length] > chunk_sz){
        NSData *chunk = [data subdataWithRange:NSMakeRange(0, chunk_sz)];
        [peripheral writeValue:chunk forCharacteristic:TxChar type:CBCharacteristicWriteWithResponse];
        sent+=chunk_sz;
        data = [data subdataWithRange:NSMakeRange(chunk_sz, [data length] - chunk_sz)];
    }
    // Send last packet (with boundary check)
    if ([data length] > 0){
        [peripheral writeValue:data forCharacteristic:TxChar type:CBCharacteristicWriteWithResponse];
        sent+=[data length];
    }
    
    NSLog(@"Wrote %ld/%ld bytes",(long)sent, (long)to_send);
    return true;
}

/**
*   Displays the specified characters on screen. If the message source is the same as the
*   previous, the characters are appended to the end of the string. Otherwise a new table
*   entry is created.
*   @param message The message to display on screen
*   @oaram characteristic The message source (TX or RX)
*/
- (void) printMessageToScreen:(NSString *)message characteristic:(CBCharacteristic *)characteristic {
    // Check if this characteristic was the last to print to screen
    if (![[direction lastObject] isEqual:characteristic]){
        // New message, check number of entries and add new one
        if ([messages count] == 20){
            [messages removeObjectAtIndex:0];
            [direction removeObjectAtIndex:0];
        }
        [direction addObject:characteristic];
        [messages addObject:message];
    } else {
        // Same source, append this to the last entry in the table
        NSString *current = [messages lastObject];
        [messages removeLastObject];
        [messages addObject:[NSString stringWithFormat:@"%@%@", current, message]];
    }
    [_messageTable reloadData];

    // Scroll to the bottom of the table so we can see the latest message
    NSInteger lastRow = [_messageTable numberOfRowsInSection:0] - 1;
    [_messageTable scrollToRowAtIndexPath:[NSIndexPath indexPathForRow:lastRow inSection:0] atScrollPosition:UITableViewScrollPositionBottom animated:NO];
}

/**
*   Converts a hex sequence to string format for display. Used when receiving hex data.
*   @param message Input sequence to be converted.
8   @return Ascii formatted sequence
*/
- (NSData *) hex2string:(NSData *)message {
    char data[message.length*3];
    unsigned char *bytePtr = (unsigned char *)[message bytes];

    int i;
    for (i=0; i<message.length*3; i+=3) {
        // Split byte into nibbles
        char b = *bytePtr++;
        char un = (b >> 4) & 0xF;
        // Ascii table 'A' offset is 0x37, '0' offset is 0x30
        data[i] = (un > 9 ? un + 0x37 : un + 0x30);
        char ln = b & 0xF;
        data[i+1] = (ln > 9 ? ln + 0x37 : ln + 0x30);
        data[i+2] = ' ';
    }
    NSData *result = [NSData dataWithBytes:&data length:i];
    return result;
}

/**
*   Converts an ascii sequence to hexadecimal ie. 'AB' -> 0xAB. The conversion will
*   fail if there are any non-hexadecimal characters present. Special charactes (e.g. spaces)
*   will be ignored. Used when transmitting hex data
*   @param message ascii formatted input sequence.
*   @return hex formatted sequence, null if invalid.
*/
- (NSData *) ascii2hex: (NSData*) message {
    unsigned char *bytePtr = (unsigned char *)[message bytes];
    unsigned char bytes[[message length]];

    // Copy the NSData array into a char array
    for (int i=0;i<[message length];i++)
        bytes[i]=*bytePtr++;

    // Store converted characters into a mutable array
    NSMutableData *in = [NSMutableData dataWithLength:0];
    for (int i=0;i<message.length;i++){
        // Check if valid hex byte
        char b = [self getValidHex:bytes[i]];
        switch (b){
            case (char)0xFF:
                // Invalid
                return NULL;
            case (char)0xFE:
                // Ignore character
                break;
            default:
                // Valid character
                [in appendBytes:&b length:1];
        }
    }
    // If there is an odd number of characters, prepend the array with '0'
    if (([in length] & 1) == 1){
        char t='0';
        NSMutableData *tmp = [NSMutableData dataWithBytes:&t length:1];
        [tmp appendData:in];
        in=tmp;
    }

    // Copy the converted sequence into a char array
    bytePtr = (unsigned char *)[in bytes];
    char inb[[in length]];
    for (int i=0;i<[in length];i++)
        inb[i]=*bytePtr++;

    // Combine the nibbles
    NSMutableData *out = [NSMutableData dataWithLength:0];
    for (int i = 0; i < [in length]/2; i++){
        char u = inb[i*2] << 4;
        char l = inb[i*2+1];
        char ul = u+l;
        [out appendBytes:&ul length:1];
    }
    // Append any EOL characters required
    char cr = 13;
    char lf = 10;
    switch (eol){
        case EOL_CR:
            [out appendBytes:&cr length:1];
            break;
        case EOL_LF:
            [out appendBytes:&lf length:1];
            break;
        case EOL_CRLF:
            [out appendBytes:&cr length:1];
            [out appendBytes:&lf length:1];
            break;
        default:
            break;
    }
    return out;

}

/**
*   Checks if a byte is a valid hexadecimal character [0-9, A-F].
*   @param b byte to check.
*   @return Ascii value of valid character, 0xFE if special character, 0xFF if invalid.
*/
- (char) getValidHex:(char) b{
    if (b<0) return 0xFE;   // Special character
    if ((b - '0') <= '9'-'0') return (b-'0');       // [0-9]
    if ((b - 'A') <= 'F'-'A') return (b-'A'+10);    // [A-F]
    if ((b - 'a') <= 'f'-'a') return (b-'a'+10);    // [a-f]
    return 0xFF;                                    // Invalid
}

/**
*   Clears the message lists and empties the message table
*/
- (void) clearScreen {
    [messages removeAllObjects];
    [direction removeAllObjects];
    [_messageTable reloadData];
}

#pragma mark UITextField

/**
*   UITextField: Called when the 'return' key is pressed on the on-screen keyboard
*   while this text field is active. Hides the on-screen keyboard.
*/
- (BOOL) textFieldShouldReturn:(UITextField *) textField {
    [textField resignFirstResponder];
    
    return true;
}

#pragma mark Button actions

/**
*   Called on 'Send' button pressed. Send any characters in the message text field.
*   Appends any EOL characters if selected.
*/
- (IBAction)onSend:(id)sender {
    NSString *message = _txt_message.text;
    
    if (message.length > 0){
        if (format == Fmt_Ascii){
            // Append any EOL characters required
            switch (eol){
                case EOL_CR:
                    message = [message stringByAppendingString:@"\r"];
                    break;
                case EOL_LF:
                    message = [message stringByAppendingString:@"\n"];
                    break;
                case EOL_CRLF:
                    message =[message stringByAppendingString:@"\r\n"];
                default:
                    break;
            }
        }
        
        if ([self writeMessageToPeripheral:message]){
            // Successfully sent, display on screen and clear text box
            [self printMessageToScreen:message characteristic:TxChar];
            _txt_message.text = @"";
        }
    }
    
    
}

- (IBAction)dismissKeyboard:(id)sender {
    
}

/**
*   Called on 'Format' button pressed. Display a popup presenting the formatting
*   choices to the user.
*/
- (IBAction)FormatButtonPressed:(id)sender {
    
    UIActionSheet *actionSheet = [[UIActionSheet alloc] initWithTitle:@"Display format"
                                                             delegate:self
                                                    cancelButtonTitle:@"Cancel"
                                               destructiveButtonTitle:nil
                                                    otherButtonTitles:@"Ascii", @"Hex", nil];
    actionSheet.tag=1;
    [actionSheet showInView:self.view];
}

/**
*   Called on 'EOL' button pressed. Display a popup presenting the EOL choices
*   to the user.
*/
- (IBAction)EOLButtonPressed:(id)sender {
    
    UIActionSheet *actionSheet = [[UIActionSheet alloc] initWithTitle:@"EOL character"
                                                             delegate:self
                                                    cancelButtonTitle:@"Cancel"
                                               destructiveButtonTitle:nil
                                                    otherButtonTitles:@"None", @"CR", @"LF", @"CR+LF", nil];
    actionSheet.tag=2;
    [actionSheet showInView:self.view];
}

/**
*   Called on 'Clear' button pressed. Clear the message display area.
*/
- (IBAction)clearScreenPressed:(id)sender {
    [self clearScreen];
}

#pragma mark Action sheet

/**
*   UIActionSheet: Delegate function called when a button is clicked on an action sheet.
*   Checks the tag and button index of the action sheet and performs the associated task.
*   @param actionSheet The action sheet identifier for the popup being displayed.
*   @param buttonIndex The identifier for the selected button.
*/
-(void)actionSheet:(UIActionSheet *)actionSheet clickedButtonAtIndex:(NSInteger)buttonIndex{
    switch (actionSheet.tag){
        // Numeric format menu
        case 1:
            NSLog(@"Format menu");
            switch (buttonIndex){
                case 0:
                    // Ascii
                    format=Fmt_Ascii;
                    break;
                case 1:
                    // Hex
                    format=Fmt_Hex;
                    break;
            }
            // Update button text
            [_Format setTitle:[NSString stringWithFormat:@"%@",format2str(format)] forState:UIControlStateNormal];
            break;
        // EOL characters menu
        case 2:
            NSLog(@"EOL menu");
            switch (buttonIndex){
                case 0:
                    // None
                    eol=EOL_NONE;
                    break;
                case 1:
                    // CR
                    eol=EOL_CR;
                    break;
                case 2:
                    // LF
                    eol=EOL_LF;
                    break;
                case 3:
                    //CR+LF
                    eol=EOL_CRLF;
                    break;
            }
            // Update button text
            [_EOL setTitle:[NSString stringWithFormat:@"EOL:%@",eol2str(eol)] forState:UIControlStateNormal];
    }
}
@end

